#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#
# Library: pydas
#
# Copyright 2010 Kitware, Inc., 28 Corporate Dr., Clifton Park, NY 12065, USA.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
###############################################################################

"""Module for the main user classes for pydas."""

import pydas.drivers
import pydas.exceptions


class Communicator(object):
    """Class for communicating with Midas Server through its drivers."""

    def __init__(self, url, drivers=None):
        """
        Constructor. Takes the URL of the Midas Server instance and an optional
        list of drivers to use.

        :param url: URL of the server
        :type url: string
        :param drivers: (optional) list of drivers to be attached to this
            communicator
        :type drivers: None | list [T <= pydas.drivers.BaseDriver]
        """
        if drivers is None:
            self._drivers = []
            import inspect

            base_driver_class = pydas.drivers.BaseDriver
            for name, obj in inspect.getmembers(pydas.drivers):
                if inspect.isclass(obj):
                    class_hierarchy = inspect.getmro(obj)
                    if base_driver_class in class_hierarchy and \
                            obj != base_driver_class:
                        instance = obj(url)
                        self._drivers.append(instance)
        else:
            self._drivers = drivers
        self._url = url

    def __getattr__(self, name):
        """
        Called when a function does not exist in the class. Pass the call down
        to one of the registered drivers.

        :raises AttributeError: if there is no function with the given name in
            any of the drivers
        """
        for driver in self.drivers:
            if hasattr(driver, name):
                return getattr(driver, name)
        raise AttributeError('{0} object has no attribute {1}'
                             .format(type(self).__name__, name))

    @property
    def drivers(self):
        """
        Get the list of drivers attached to this communicator.

        :returns: list of drivers
        :rtype: list[T <= pydas.drivers.BaseDriver]
        """
        return self._drivers

    @property
    def url(self):
        """
        Return the URL of the server.

        :returns: URL of the server
        :rtype: string
        """
        if len(self.drivers) > 0:
            return self.drivers[0].url
        else:
            return self._url

    @url.setter
    def url(self, value):
        """
        Set the URL of the server in all drivers attached to this communicator.

        :param value: URL of the server
        :type value: string
        """
        for driver in self.drivers:
            driver.url = value

    @property
    def debug(self):
        """
        Return whether the debug state of every driver is True.

        :returns: True if the debug state of every driver is True
        :rtype: bool
        """
        return all(driver.debug for driver in self.drivers)

    @debug.setter
    def debug(self, value):
        """
        Set the debug state of all of drivers attached to this communicator.

        :param value: debug state of all drivers
        :type value: bool
        """
        for driver in self.drivers:
            driver.debug = value

    @property
    def verify_ssl_certificate(self):
        """
        Return whether the SSL certificate will be verified for all drivers
        attached to this communicator.

        :returns: True if the SSL certificate will be verified for every driver
        :rtype: bool
        """
        return all(driver.verify_ssl_certificate for driver in self.drivers)

    @verify_ssl_certificate.setter
    def verify_ssl_certificate(self, value):
        """
        Set whether the SSL certificate will be verified.

        :param value: If True, the SSL certificate will be verified
        :type value: bool
        """
        for driver in self.drivers:
            driver.verify_ssl_certificate = value

    def set_auth(self, value):
        """
        Set the authentication in all drivers attached to this communicator.

        :param value: authentication tuple to be passed to requests.request()
        :type value: None | tuple
        """
        for driver in self.drivers:
            driver.auth = value



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#
# Library: pydas
#
# Copyright 2010 Kitware, Inc., 28 Corporate Dr., Clifton Park, NY 12065, USA.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
###############################################################################

"""API for pydas."""

import getpass
import glob
import os
import os.path
import hashlib

from core import Communicator
import session


def login(email=None, password=None, api_key=None, application='Default',
          url=None, verify_ssl_certificate=True):
    """
    Do the legwork of logging into the Midas Server instance, storing the API
    key and token.

    :param email: (optional) Email address to login with. If not set, the
        console will be prompted.
    :type email: None | string
    :param password: (optional) User password to login with. If not set and no
        'api_key' is set, the console will be prompted.
    :type password: None | string
    :param api_key: (optional) API key to login with. If not set, password
        login with be used.
    :type api_key: None | string
    :param application: (optional) Application name to be used with 'api_key'.
    :type application: string
    :param url: (optional) URL address of the Midas Server instance to login
        to. If not set, the console will be prompted.
    :type url: None | string
    :param verify_ssl_certificate: (optional) If True, the SSL certificate will
        be verified
    :type verify_ssl_certificate: bool
    :returns: API token.
    :rtype: string
    """
    try:
        input_ = raw_input
    except NameError:
        input_ = input

    if url is None:
        url = input_('Server URL: ')
    url = url.rstrip('/')
    if session.communicator is None:
        session.communicator = Communicator(url)
    else:
        session.communicator.url = url

    session.communicator.verify_ssl_certificate = verify_ssl_certificate

    if email is None:
        email = input_('Email: ')
    session.email = email

    if api_key is None:
        if password is None:
            password = getpass.getpass()
        session.api_key = session.communicator.get_default_api_key(
            session.email, password)
        session.application = 'Default'
    else:
        session.api_key = api_key
        session.application = application

    return renew_token()


def renew_token():
    """
    Renew or get a token to use for transactions with the Midas Server
    instance.

    :returns: API token.
    :rtype: string
    """
    session.token = session.communicator.login_with_api_key(
        session.email, session.api_key, application=session.application)
    if len(session.token) < 10:  # HACK to check for mfa being enabled
        one_time_pass = getpass.getpass('One-Time Password: ')
        session.token = session.communicator.mfa_otp_login(
            session.token, one_time_pass)
    return session.token


def verify_credentials():
    """
    Check if the current credentials are valid and login or renew as needed.

    :returns: API token.
    :rtype: string
    """
    if session.token is not None:
        return session.token
    elif session.api_key is not None:
        return renew_token()
    else:
        return login()


def add_item_upload_callback(callback):
    """
    Pass a function to be called when an item is created. This can be quite
    useful for performing actions such as notifications of upload progress as
    well as calling additional API functions.

    :param callback: A function that takes three arguments. The first argument
        is the communicator object of the current pydas context, the second is
        the currently active API token and the third is the id of the item that
        was created to result in the callback function's invocation.
    :type callback: (Communicator, string, int) -> unknown
    """
    session.item_upload_callbacks.append(callback)


def _create_or_reuse_item(local_file, parent_folder_id, reuse_existing=False):
    """
    Create an item from the local file in the Midas Server folder corresponding
    to the parent folder id.

    :param local_file: full path to a file on the local file system
    :type local_file: string
    :param parent_folder_id: id of parent folder on the Midas Server instance,
        where the item will be added
    :type parent_folder_id: int | long
    :param reuse_existing: (optional) whether to accept an existing item of the
        same name in the same location, or create a new one instead
    :type reuse_existing: bool
    """
    local_item_name = os.path.basename(local_file)
    item_id = None
    if reuse_existing:
        # check by name to see if the item already exists in the folder
        children = session.communicator.folder_children(
            session.token, parent_folder_id)
        items = children['items']

        for item in items:
            if item['name'] == local_item_name:
                item_id = item['item_id']
                break

    if item_id is None:
        # create the item for the subdir
        new_item = session.communicator.create_item(
            session.token, local_item_name, parent_folder_id)
        item_id = new_item['item_id']

    return item_id


def _create_or_reuse_folder(local_folder, parent_folder_id,
                            reuse_existing=False):
    """
    Create a folder from the local file in the midas folder corresponding to
    the parent folder id.

    :param local_folder: full path to a directory on the local file system
    :type local_folder: string
    :param parent_folder_id: id of parent folder on the Midas Server instance,
        where the folder will be added
    :type parent_folder_id: int | long
    :param reuse_existing: (optional) whether to accept an existing folder of
       the same name in the same location, or create a new one instead
    :type reuse_existing: bool
    """
    local_folder_name = os.path.basename(local_folder)
    folder_id = None
    if reuse_existing:
        # check by name to see if the folder already exists in the folder
        children = session.communicator.folder_children(
            session.token, parent_folder_id)
        folders = children['folders']

        for folder in folders:
            if folder['name'] == local_folder_name:
                folder_id = folder['folder_id']
                break

    if folder_id is None:
        # create the item for the subdir
        new_folder = session.communicator.create_folder(session.token,
                                                        local_folder_name,
                                                        parent_folder_id)
        folder_id = new_folder['folder_id']

    return folder_id


def _streaming_file_md5(file_path):
    """
    Create and return a hex checksum using the MD5 sum of the passed in file.
    This will stream the file, rather than load it all into memory.

    :param file_path: full path to the file
    :type file_path: string
    :returns: a hex checksum
    :rtype: string
    """
    md5 = hashlib.md5()
    with open(file_path, 'rb') as f:
        # iter needs an empty byte string for the returned iterator to halt at
        # EOF
        for chunk in iter(lambda: f.read(128 * md5.block_size), b''):
            md5.update(chunk)
    return md5.hexdigest()


def _create_bitstream(file_path, local_file, item_id, log_ind=None):
    """
    Create a bitstream in the given item.

    :param file_path: full path to the local file
    :type file_path: string
    :param local_file: name of the local file
    :type local_file: string
    :param log_ind: (optional) any additional message to log upon creation of
        the bitstream
    :type log_ind: None | string
    """
    checksum = _streaming_file_md5(file_path)
    upload_token = session.communicator.generate_upload_token(
        session.token, item_id, local_file, checksum)

    if upload_token != '':
        log_trace = 'Uploading bitstream from {0}'.format(file_path)
        # only need to perform the upload if we haven't uploaded before
        # in this cae, the upload token would not be empty
        session.communicator.perform_upload(
            upload_token, local_file, filepath=file_path, itemid=item_id)
    else:
        log_trace = 'Adding a bitstream link in this item to an existing ' \
                    'bitstream from {0}'.format(file_path)

    if log_ind is not None:
        log_trace += log_ind
    print(log_trace)


def _upload_as_item(local_file, parent_folder_id, file_path,
                    reuse_existing=False):
    """
    Function for doing an upload of a file as an item. This should be a
    building block for user-level functions.

    :param local_file: name of local file to upload
    :type local_file: string
    :param parent_folder_id: id of parent folder on the Midas Server instance,
        where the item will be added
    :type parent_folder_id: int | long
    :param file_path: full path to the file
    :type file_path: string
    :param reuse_existing: (optional) whether to accept an existing item of the
        same name in the same location, or create a new one instead
    :type reuse_existing: bool
    """
    current_item_id = _create_or_reuse_item(local_file, parent_folder_id,
                                            reuse_existing)
    _create_bitstream(file_path, local_file, current_item_id)
    for callback in session.item_upload_callbacks:
        callback(session.communicator, session.token, current_item_id)


def _create_folder(local_folder, parent_folder_id):
    """
    Function for creating a remote folder and returning the id. This should be
    a building block for user-level functions.

    :param local_folder: full path to a local folder
    :type local_folder: string
    :param parent_folder_id: id of parent folder on the Midas Server instance,
        where the new folder will be added
    :type parent_folder_id: int | long
    :returns: id of the remote folder that was created
    :rtype: int | long
    """
    new_folder = session.communicator.create_folder(
        session.token, os.path.basename(local_folder), parent_folder_id)
    return new_folder['folder_id']


def _upload_folder_recursive(local_folder,
                             parent_folder_id,
                             leaf_folders_as_items=False,
                             reuse_existing=False):
    """
    Function to recursively upload a folder and all of its descendants.

    :param local_folder: full path to local folder to be uploaded
    :type local_folder: string
    :param parent_folder_id: id of parent folder on the Midas Server instance,
        where the new folder will be added
    :type parent_folder_id: int | long
    :param leaf_folders_as_items: (optional) whether leaf folders should have
        all files uploaded as single items
    :type leaf_folders_as_items: bool
    :param reuse_existing: (optional) whether to accept an existing item of the
        same name in the same location, or create a new one instead
    :type reuse_existing: bool
    """
    if leaf_folders_as_items and _has_only_files(local_folder):
        print('Creating item from {0}'.format(local_folder))
        _upload_folder_as_item(local_folder, parent_folder_id, reuse_existing)
        return
    else:
        # do not need to check if folder exists, if it does, an attempt to
        # create it will just return the existing id
        print('Creating folder from {0}'.format(local_folder))
        new_folder_id = _create_or_reuse_folder(local_folder, parent_folder_id,
                                                reuse_existing)

        for entry in sorted(os.listdir(local_folder)):
            full_entry = os.path.join(local_folder, entry)
            if os.path.islink(full_entry):
                # os.walk skips symlinks by default
                continue
            elif os.path.isdir(full_entry):
                _upload_folder_recursive(full_entry,
                                         new_folder_id,
                                         leaf_folders_as_items,
                                         reuse_existing)
            else:
                print('Uploading item from {0}'.format(full_entry))
                _upload_as_item(entry,
                                new_folder_id,
                                full_entry,
                                reuse_existing)


def _has_only_files(local_folder):
    """
    Return whether a folder contains only files. This will be False if the
    folder contains any subdirectories.

    :param local_folder: full path to the local folder
    :type local_folder: string
    :returns: True if the folder contains only files
    :rtype: bool
    """
    return not any(os.path.isdir(os.path.join(local_folder, entry))
                   for entry in os.listdir(local_folder))


def _upload_folder_as_item(local_folder, parent_folder_id,
                           reuse_existing=False):
    """
    Upload a folder as a new item. Take a folder and use its base name as the
    name of a new item. Then, upload its containing files into the new item as
    bitstreams.

    :param local_folder: The path to the folder to be uploaded
    :type local_folder: string
    :param parent_folder_id: The id of the destination folder for the new item.
    :type parent_folder_id: int | long
    :param reuse_existing: (optional) whether to accept an existing item of the
        same name in the same location, or create a new one instead
    :type reuse_existing: bool
    """
    item_id = _create_or_reuse_item(local_folder, parent_folder_id,
                                    reuse_existing)

    subdir_contents = sorted(os.listdir(local_folder))
    # for each file in the subdir, add it to the item
    filecount = len(subdir_contents)
    for (ind, current_file) in enumerate(subdir_contents):
        file_path = os.path.join(local_folder, current_file)
        log_ind = '({0} of {1})'.format(ind + 1, filecount)
        _create_bitstream(file_path, current_file, item_id, log_ind)

    for callback in session.item_upload_callbacks:
        callback(session.communicator, session.token, item_id)


def upload(file_pattern, destination='Private', leaf_folders_as_items=False,
           reuse_existing=False):
    """
    Upload a pattern of files. This will recursively walk down every tree in
    the file pattern to create a hierarchy on the server. As of right now, this
    places the file into the currently logged in user's home directory.

    :param file_pattern: a glob type pattern for files
    :type file_pattern: string
    :param destination: (optional) name of the midas destination folder,
        defaults to Private
    :type destination: string
    :param leaf_folders_as_items: (optional) whether leaf folders should have
        all files uploaded as single items
    :type leaf_folders_as_items: bool
    :param reuse_existing: (optional) whether to accept an existing item of the
        same name in the same location, or create a new one instead
    :type reuse_existing: bool
    """
    session.token = verify_credentials()

    # Logic for finding the proper folder to place the files in.
    parent_folder_id = None
    user_folders = session.communicator.list_user_folders(session.token)
    if destination.startswith('/'):
        parent_folder_id = _find_resource_id_from_path(destination)
    else:
        for cur_folder in user_folders:
            if cur_folder['name'] == destination:
                parent_folder_id = cur_folder['folder_id']
    if parent_folder_id is None:
        print('Unable to locate specified destination. Defaulting to {0}.'
              .format(user_folders[0]['name']))
        parent_folder_id = user_folders[0]['folder_id']

    for current_file in glob.iglob(file_pattern):
        current_file = os.path.normpath(current_file)
        if os.path.isfile(current_file):
            print('Uploading item from {0}'.format(current_file))
            _upload_as_item(os.path.basename(current_file),
                            parent_folder_id,
                            current_file,
                            reuse_existing)
        else:
            _upload_folder_recursive(current_file,
                                     parent_folder_id,
                                     leaf_folders_as_items,
                                     reuse_existing)


def _descend_folder_for_id(parsed_path, folder_id):
    """
    Descend a path to return a folder id starting from the given folder id.

    :param parsed_path: a list of folders from top to bottom of a hierarchy
    :type parsed_path: list[string]
    :param folder_id: The id of the folder from which to start the descent
    :type folder_id: int | long
    :returns: The id of the found folder or -1
    :rtype: int | long
    """
    if len(parsed_path) == 0:
        return folder_id

    session.token = verify_credentials()

    base_folder = session.communicator.folder_get(session.token,
                                                  folder_id)
    cur_folder_id = -1
    for path_part in parsed_path:
        cur_folder_id = base_folder['folder_id']
        cur_children = session.communicator.folder_children(
            session.token, cur_folder_id)
        for inner_folder in cur_children['folders']:
            if inner_folder['name'] == path_part:
                base_folder = session.communicator.folder_get(
                    session.token, inner_folder['folder_id'])
                cur_folder_id = base_folder['folder_id']
                break
        else:
            return -1
    return cur_folder_id


def _search_folder_for_item_or_folder(name, folder_id):
    """
    Find an item or folder matching the name. A folder will be found first if
    both are present.

    :param name: The name of the resource
    :type name: string
    :param folder_id: The folder to search within
    :type folder_id: int | long
    :returns: A tuple indicating whether the resource is an item an the id of
        said resource. i.e. (True, item_id) or (False, folder_id). Note that in
        the event that we do not find a result return (False, -1)
    :rtype: (bool, int | long)
    """
    session.token = verify_credentials()

    children = session.communicator.folder_children(session.token, folder_id)
    for folder in children['folders']:
        if folder['name'] == name:
            return False, folder['folder_id']  # Found a folder
    for item in children['items']:
        if item['name'] == name:
            return True, item['item_id']  # Found an item
    return False, -1  # Found nothing


def _find_resource_id_from_path(path):
    """
    Get a folder id from a path on the server.

    Warning: This is NOT efficient at all.

    The schema for this path is:
    path := "/users/<name>/" | "/communities/<name>" , {<subfolder>/}
    name := <firstname> , "_" , <lastname>

    :param path: The virtual path on the server.
    :type path: string
    :returns: a tuple indicating True or False about whether the resource is an
        item and id of the resource i.e. (True, item_id) or (False, folder_id)
    :rtype: (bool, int | long)
    """
    session.token = verify_credentials()

    parsed_path = path.split('/')
    if parsed_path[-1] == '':
        parsed_path.pop()
    if path.startswith('/users/'):
        parsed_path.pop(0)  # remove '' before /
        parsed_path.pop(0)  # remove 'users'
        name = parsed_path.pop(0)  # remove '<firstname>_<lastname>'
        firstname, lastname = name.split('_')
        end = parsed_path.pop()
        user = session.communicator.get_user_by_name(firstname, lastname)
        leaf_folder_id = _descend_folder_for_id(parsed_path, user['folder_id'])
        return _search_folder_for_item_or_folder(end, leaf_folder_id)
    elif path.startswith('/communities/'):
        print(parsed_path)
        parsed_path.pop(0)  # remove '' before /
        parsed_path.pop(0)  # remove 'communities'
        community_name = parsed_path.pop(0)  # remove '<community>'
        end = parsed_path.pop()
        community = session.communicator.get_community_by_name(community_name)
        leaf_folder_id = _descend_folder_for_id(parsed_path,
                                                community['folder_id'])
        return _search_folder_for_item_or_folder(end, leaf_folder_id)
    else:
        return False, -1


def _download_folder_recursive(folder_id, path='.'):
    """
    Download a folder to the specified path along with any children.

    :param folder_id: The id of the target folder
    :type folder_id: int | long
    :param path: (optional) the location to download the folder
    :type path: string
    """
    session.token = verify_credentials()

    cur_folder = session.communicator.folder_get(session.token, folder_id)
    folder_path = os.path.join(path, cur_folder['name'])
    print('Creating folder at {0}'.format(folder_path))
    os.mkdir(folder_path)
    cur_children = session.communicator.folder_children(
        session.token, folder_id)
    for item in cur_children['items']:
        _download_item(item['item_id'], folder_path)
    for folder in cur_children['folders']:
        _download_folder_recursive(folder['folder_id'], folder_path)


def _download_item(item_id, path='.'):
    """
    Download the requested item to the specified path.

    :param item_id: The id of the item to be downloaded
    :type item_id: int | long
    :param path: (optional) the location to download the item
    :type path: string
    """
    session.token = verify_credentials()

    filename, content_iter = session.communicator.download_item(
        item_id, session.token)
    item_path = os.path.join(path, filename)
    print('Creating file at {0}'.format(item_path))
    out_file = open(item_path, 'wb')
    for block in content_iter:
        out_file.write(block)
    out_file.close()


def download(server_path, local_path='.'):
    """
    Recursively download a file or item from the Midas Server instance.

    :param server_path: The location on the server to find the resource to
        download
    :type server_path: string
    :param local_path: The location on the client to store the downloaded data
    :type local_path: string
    """
    session.token = verify_credentials()

    is_item, resource_id = _find_resource_id_from_path(server_path)
    if resource_id == -1:
        print('Unable to locate {0}'.format(server_path))
    else:
        if is_item:
            _download_item(resource_id, local_path)
        else:
            _download_folder_recursive(resource_id, local_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#
# Library: pydas
#
# Copyright 2010 Kitware, Inc., 28 Corporate Dr., Clifton Park, NY 12065, USA.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
###############################################################################

"""Module for exceptions in pydas."""

import requests.exceptions


class PydasException(Exception):
    """Base class for exceptions defined in pydas."""

    def __init__(self, value):
        """
        Override the constructor to support a basic message.

        :param value: Message to display.
        :type value: string
        """
        super(PydasException, self).__init__()
        self.value = value
        self.method = None
        self.code = None

    def __str__(self):
        """
        Override the string method.

        :returns: String representation of the message.
        :rtype: string
        """
        return repr(self.value)


class RequestError(PydasException, requests.exceptions.ConnectionError):
    """Error relating to an HTTP request."""

    def __init__(self, value, request=None):
        """
        Override the constructor to store the request sent.

        :param value: Message to display.
        :type value: string
        :param request: (optional) Request sent.
        :type request: requests.Request
        """
        super(RequestError, self).__init__(value)
        self.request = request


class SSLVerificationFailed(RequestError, requests.exceptions.SSLError):
    """Error verifying an SSL certificate."""


class ResponseError(PydasException):
    """Error relating to an HTTP response."""

    def __init__(self, value, response=None):
        """
        Override the constructor to store the response received.

        :param value: Message to display.
        :type value: string
        :param response: (optional) Response received.
        :type response: requests.Response
        """
        super(ResponseError, self).__init__(value)
        self.response = response


class ParseError(ResponseError, ValueError):
    """Error parsing a response."""


class HTTPError(ResponseError, requests.exceptions.HTTPError):
    """HTTP status code 400 or above."""


class BadRequest(HTTPError):
    """HTTP status code 400: bad request."""


class InternalError(BadRequest):
    """Midas Server error code -100: internal error."""


class InvalidParameter(BadRequest, ValueError):
    """Midas Server error code -150: internal error."""


class InvalidPolicy(BadRequest, ValueError):
    """Midas Server error code  -151: invalid policy."""


class UploadFailed(BadRequest):
    """Midas Server error code -105: upload failed."""


class UploadTokenGenerationFailed(BadRequest):
    """Midas Server error code -140: upload token generation failed."""


class Unauthorized(HTTPError):
    """HTTP status code 401: unauthorized."""


class InvalidToken(Unauthorized, ValueError):
    """Midas Server error code -101: invalid token."""


class InvalidUploadToken(Unauthorized, ValueError):
    """Midas Server error code -141: invalid upload token."""


class Forbidden(HTTPError):
    """HTTP status code 403: forbidden."""


class NotFound(HTTPError):
    """HTTP status code 404: not found or HTTP status code 410: gone."""


class MethodNotAllowed(HTTPError):
    """HTTP status code 405: method not allowed."""


def get_exception_from_status_and_error_codes(status_code, error_code, value):
    """
    Return an exception given status and error codes.

    :param status_code: HTTP status code.
    :type status_code: None | int
    :param error_code: Midas Server error code.
    :type error_code: None | int
    :param value: Message to display.
    :type value: string
    :returns: Exception.
    :rtype : pydas.exceptions.ResponseError
    """
    if status_code == requests.codes.bad_request:
        exception = BadRequest(value)
    elif status_code == requests.codes.unauthorized:
        exception = Unauthorized(value)
    elif status_code == requests.codes.forbidden:
        exception = Unauthorized(value)
    elif status_code in [requests.codes.not_found, requests.codes.gone]:
        exception = NotFound(value)
    elif status_code == requests.codes.method_not_allowed:
        exception = MethodNotAllowed(value)
    elif status_code >= requests.codes.bad_request:
        exception = HTTPError(value)
    else:
        exception = ResponseError(value)

    if error_code == -100:  # MIDAS_INTERNAL_ERROR
        exception = InternalError(value)
    elif error_code == -101:  # MIDAS_INVALID_TOKEN
        exception = InvalidToken(value)
    elif error_code == -105:  # MIDAS_UPLOAD_FAILED
        exception = UploadFailed(value)
    elif error_code == -140:  # MIDAS_UPLOAD_TOKEN_GENERATION_FAILED
        exception = UploadTokenGenerationFailed(value)
    elif error_code == -141:  # MIDAS_INVALID_UPLOAD_TOKEN
        exception = InvalidUploadToken(value)
    elif error_code == -150:  # MIDAS_INVALID_PARAMETER
        exception = InvalidParameter(value)
    elif error_code == -151:  # MIDAS_INVALID_POLICY
        exception = InvalidPolicy(value)

    return exception



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#
# Library: pydas
#
# Copyright 2010 Kitware, Inc., 28 Corporate Dr., Clifton Park, NY 12065, USA.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
###############################################################################

"""
This module is for the drivers that actually do the work of communication
with the Midas Server instance. Any drivers that are implemented should use the
utility functions provided in pydas.drivers.BaseDriver by inheriting from that
class.
"""

import json
import os

import requests
import requests.exceptions

from retrying import retry

import pydas.exceptions


class BaseDriver(object):
    """Base class for the Midas Server API drivers."""

    # Class members.
    email = ''
    apikey = ''

    def __init__(self, url=""):
        """
        Constructor.

        :param url: (optional) URL of the server
        :type url: string
        """
        self._api_suffix = '/api/json?method='
        self._url = url
        self._debug = False
        self._verify_ssl_certificate = True
        self.auth = None

    @property
    def url(self):
        """
        Return the URL of the server.

        :returns: URL of the server
        :rtype: string
        """
        return self._url

    @url.setter
    def url(self, value):
        """
        Set the URL of the server.

        :param value: URL of the server
        :type value: string
        """
        self._url = value

    @property
    def full_url(self):
        """
        Return the full URL of the server including the API suffix.

        :returns: Full URL of the server
        :rtype: string
        """
        return self._url + self._api_suffix

    @property
    def debug(self):
        """
        Return the debug state of this driver.

        :returns: debug state
        :rtype: bool
        """
        return self._debug

    @debug.setter
    def debug(self, value):
        """
        Set the debug state of this driver.

        :param value: debug state
        :type value: bool
        """
        self._debug = value

    @property
    def verify_ssl_certificate(self):
        """
        Return whether the SSL certificate will be verified.

        :returns: True if the SSL certificate will be verified
        :rtype: bool
        """
        return self._verify_ssl_certificate

    @verify_ssl_certificate.setter
    def verify_ssl_certificate(self, value):
        """
        Set whether the SSL certificate will be verified.

        :param value: If True, the SSL certificate will be verified
        :type value: bool
        """
        self._verify_ssl_certificate = value

    @retry(wait_exponential_multiplier=1000,
           wait_exponential_max=5000,
           stop_max_attempt_number=5)
    def request(self, method, parameters=None, file_payload=None):
        """
        Do the generic processing of a request to the server.

        If file_payload is specified, it will be PUT to the server.

        :param method: Desired API method
        :type method: string
        :param parameters: (optional) Parameters to pass in the HTTP body
        :type parameters: None | dict[string, string]
        :param file_payload: (optional) File-like object to be sent with the
            HTTP request
        :type file_payload: None | file | FileIO
        :returns: Dictionary representing the JSON response to the request
        :rtype: dict
        :raises pydas.exceptions.PydasException: if the request failed
        """
        method_url = self.full_url + method
        response = None

        try:
            if file_payload:
                response = requests.put(method_url,
                                        data=file_payload.read(),
                                        params=parameters,
                                        allow_redirects=True,
                                        verify=self._verify_ssl_certificate,
                                        auth=self.auth)
            else:
                response = requests.post(method_url,
                                         params=parameters,
                                         allow_redirects=True,
                                         verify=self._verify_ssl_certificate,
                                         auth=self.auth)

        except requests.exceptions.SSLError:
            exception = pydas.exceptions.SSLVerificationFailed(
                'Request failed with an SSL verification error')
            exception.method = method
            exception.request = response.request
            raise exception

        except requests.exceptions.ConnectionError:
            exception = pydas.exceptions.RequestError(
                'Request failed with a connection error')
            exception.method = method
            if response is not None:
                exception.request = response.request
            raise exception

        status_code = response.status_code

        try:
            response.raise_for_status()

        except requests.exceptions.HTTPError:
            error_code = None
            message = 'Request failed with HTTP status code {0}'.format(
                status_code)

            try:
                content = response.json()

                if 'code' in content:
                    error_code = int(content['code'])
                    message = 'Request failed with HTTP status code {0}, ' \
                        'Midas Server error code {1}, and response content ' \
                        '{2}'.format(status_code, error_code, response.content)

            except ValueError:
                pass

            exception = pydas.exceptions \
                .get_exception_from_status_and_error_codes(status_code,
                                                           error_code,
                                                           message)
            exception.code = error_code
            exception.method = method
            exception.response = response
            raise exception

        try:
            content = response.json()

        except ValueError:
            exception = pydas.exceptions.ParseError(
                'Request failed with HTTP status code {0} and response '
                'content {1}'.format(status_code, response.content))
            exception.method = method
            exception.response = response
            raise exception

        if 'stat' not in content:
            exception = pydas.exceptions.ParseError(
                'Request failed with HTTP status code {0} and response '
                'content {1}'.format(status_code, response.content))
            exception.method = method
            raise exception

        if content['stat'] != 'ok':
            if 'code' in content:
                error_code = int(content['code'])
                message = 'Request failed with HTTP status code {0}, Midas ' \
                    'Server error code {1}, and response content {2}' \
                    .format(status_code, error_code, response.content)
            else:
                error_code = None
                message = 'Request failed with HTTP status code {0} and ' \
                    'response content {1}'.format(status_code,
                                                  response.content)

            exception = pydas.exceptions \
                .get_exception_from_status_and_error_codes(status_code,
                                                           error_code,
                                                           message)
            exception.method = method
            exception.response = response
            raise exception

        if 'data' not in content:
            exception = pydas.exceptions.ParseError(
                'Request failed with HTTP status code {0} and response '
                'content {1}'.format(status_code, response.content))
            exception.method = method
            exception.response = response
            raise exception

        return content['data']

    def login_with_api_key(self, email, api_key, application='Default'):
        """
        Login and get a token. If you do not specify a specific application,
        'Default' will be used.

        :param email: Email address of the user
        :type email: string
        :param api_key: API key assigned to the user
        :type api_key: string
        :param application: (optional) Application designated for this API key
        :type application: string
        :returns: Token to be used for interaction with the API until
            expiration
        :rtype: string
        """
        parameters = dict()
        parameters['email'] = BaseDriver.email = email  # Cache email
        parameters['apikey'] = BaseDriver.apikey = api_key  # Cache API key
        parameters['appname'] = application
        response = self.request('midas.login', parameters)
        if 'token' in response:  # normal case
            return response['token']
        if 'mfa_token_id':  # case with multi-factor authentication
            return response['mfa_token_id']


class CoreDriver(BaseDriver):
    """
    Driver for the core API methods of Midas Server. This contains all of the
    calls necessary to interact with a Midas Server instance that has no
    plugins enabled (other than the web API).
    """

    def get_server_version(self):
        """
        Get the version from the server.

        :returns: version code from the server
        :rtype: string
        """
        response = self.request('midas.version')
        return response['version']

    def get_server_info(self):
        """
        Get general server information.

        The information provided includes enabled modules as well as enabled
        web API functions.

        :returns: Module and web API information.
        :rtype: dict
        """
        response = self.request('midas.info')
        return response['version']

    def list_modules(self):
        """
        List the enabled modules on the server.

        :returns: List of names of the enabled modules.
        :rtype: list[string]
        """
        response = self.request('midas.modules.list')
        return response['modules']

    def list_user_folders(self, token):
        """
        List the folders in the users home area.

        :param token: A valid token for the user in question.
        :type token: string
        :returns: List of dictionaries containing folder information.
        :rtype: list[dict]
        """
        parameters = dict()
        parameters['token'] = token
        response = self.request('midas.user.folders', parameters)
        return response

    def get_default_api_key(self, email, password):
        """
        Get the default API key for a user.

        :param email: The email of the user.
        :type email: string
        :param password: The user's password.
        :type password: string
        :returns: API key to confirm that it was fetched successfully.
        :rtype: string
        """
        parameters = dict()
        parameters['email'] = email
        parameters['password'] = password
        response = self.request('midas.user.apikey.default', parameters)
        return response['apikey']

    def list_users(self, limit=20):
        """
        List the public users in the system.

        :param limit: (optional) The number of users to fetch.
        :type limit: int | long
        :returns: The list of users.
        :rtype: list[dict]
        """
        parameters = dict()
        parameters['limit'] = limit
        response = self.request('midas.user.list', parameters)
        return response

    def get_user_by_name(self, firstname, lastname):
        """
        Get a user by the first and last name of that user.

        :param firstname: The first name of the user.
        :type firstname: string
        :param lastname: The last name of the user.
        :type lastname: string
        :returns: The user requested.
        :rtype: dict
        """
        parameters = dict()
        parameters['firstname'] = firstname
        parameters['lastname'] = lastname
        response = self.request('midas.user.get', parameters)
        return response

    def get_user_by_id(self, user_id):
        """
        Get a user by the first and last name of that user.

        :param user_id: The id of the desired user.
        :type user_id: int | long
        :returns: The user requested.
        :rtype: dict
        """
        parameters = dict()
        parameters['user_id'] = user_id
        response = self.request('midas.user.get', parameters)
        return response

    def get_user_by_email(self, email):
        """
        Get a user by the email of that user.

        :param email: The email of the desired user.
        :type email: string
        :returns: The user requested.
        :rtype: dict
        """
        parameters = dict()
        parameters['email'] = email
        response = self.request('midas.user.get', parameters)
        return response

    def create_community(self, token, name, **kwargs):
        """
        Create a new community or update an existing one using the uuid.

        :param token: A valid token for the user in question.
        :type token: string
        :param name: The community name.
        :type name: string
        :param description: (optional) The community description.
        :type description: string
        :param uuid: (optional) uuid of the community. If none is passed, will
            generate one.
        :type uuid: string
        :param privacy: (optional) Default 'Public', possible values
            [Public|Private].
        :type privacy: string
        :param can_join: (optional) Default 'Everyone', possible values
            [Everyone|Invitation].
        :type can_join: string
        :returns: The community dao that was created.
        :rtype: dict
        """
        parameters = dict()
        parameters['token'] = token
        parameters['name'] = name
        optional_keys = ['description', 'uuid', 'privacy', 'can_join']
        for key in optional_keys:
            if key in kwargs:
                if key == 'can_join':
                    parameters['canjoin'] = kwargs[key]
                    continue
                parameters[key] = kwargs[key]
        response = self.request('midas.community.create', parameters)
        return response

    def get_community_by_name(self, name, token=None):
        """
        Get a community based on its name.

        :param name: The name of the target community.
        :type name: string
        :param token: (optional) A valid token for the user in question.
        :type token: None | string
        :returns: The requested community.
        :rtype: dict
        """
        parameters = dict()
        parameters['name'] = name
        if token:
            parameters['token'] = token
        response = self.request('midas.community.get', parameters)
        return response

    def get_community_by_id(self, community_id, token=None):
        """
        Get a community based on its id.

        :param community_id: The id of the target community.
        :type community_id: int | long
        :param token: (optional) A valid token for the user in question.
        :type token: None | string
        :returns: The requested community.
        :rtype: dict
        """
        parameters = dict()
        parameters['id'] = community_id
        if token:
            parameters['token'] = token
        response = self.request('midas.community.get', parameters)
        return response

    def get_community_children(self, community_id, token=None):
        """
        Get the non-recursive children of the passed in community_id.

        :param community_id: The id of the requested community.
        :type community_id: int | long
        :param token: (optional) A valid token for the user in question.
        :type token: None | string
        :returns: List of the folders in the community.
        :rtype: dict[string, list]
        """
        parameters = dict()
        parameters['id'] = community_id
        if token:
            parameters['token'] = token
        response = self.request('midas.community.children', parameters)
        return response

    def list_communities(self, token=None):
        """
        List all communities visible to a user.

        :param token: (optional) A valid token for the user in question.
        :type token: None | string
        :returns: The list of communities.
        :rtype: list[dict]
        """
        parameters = dict()
        if token:
            parameters['token'] = token
        response = self.request('midas.community.list', parameters)
        return response

    def create_folder(self, token, name, parent_id, **kwargs):
        """
        Create a folder at the destination specified.

        :param token: A valid token for the user in question.
        :type token: string
        :param name: The name of the folder to be created.
        :type name: string
        :param parent_id: The id of the targeted parent folder.
        :type parent_id: int | long
        :param description: (optional) The description text of the folder.
        :type description: string
        :param uuid: (optional) The UUID for the folder. It will be generated
            if not given.
        :type uuid: string
        :param privacy: (optional) The privacy state of the folder
            ('Public' or 'Private').
        :param reuse_existing: (optional) If true, will just return the
            existing folder if there is one with the name provided.
        :type reuse_existing: bool
        :returns: Dictionary containing the details of the created folder.
        :rtype: dict
        """
        parameters = dict()
        parameters['token'] = token
        parameters['name'] = name
        parameters['parentid'] = parent_id
        parameters['description'] = ''
        optional_keys = ['description', 'uuid', 'privacy', 'reuse_existing']
        for key in optional_keys:
            if key in kwargs:
                if key == 'reuse_existing':
                    if kwargs[key]:
                        parameters['reuseExisting'] = kwargs[key]
                    continue
                parameters[key] = kwargs[key]
        response = self.request('midas.folder.create', parameters)
        return response

    def folder_get(self, token, folder_id):
        """
        Get the attributes of the specified folder.

        :param token: A valid token for the user in question.
        :type token: string
        :param folder_id: The id of the requested folder.
        :type folder_id: int | long
        :returns: Dictionary of the folder attributes.
        :rtype: dict
        """
        parameters = dict()
        parameters['token'] = token
        parameters['id'] = folder_id
        response = self.request('midas.folder.get', parameters)
        return response

    def folder_children(self, token, folder_id):
        """
        Get the non-recursive children of the passed in folder_id.

        :param token: A valid token for the user in question.
        :type token: string
        :param folder_id: The id of the requested folder.
        :type folder_id: int | long
        :returns: Dictionary of two lists: 'folders' and 'items'.
        :rtype: dict[string, list]
        """
        parameters = dict()
        parameters['token'] = token
        parameters['id'] = folder_id
        response = self.request('midas.folder.children', parameters)
        return response

    def delete_folder(self, token, folder_id):
        """
        Delete the folder with the passed in folder_id.

        :param token: A valid token for the user in question.
        :type token: string
        :param folder_id: The id of the folder to be deleted.
        :type folder_id: int | long
        :returns: None.
        :rtype: None
        """
        parameters = dict()
        parameters['token'] = token
        parameters['id'] = folder_id
        response = self.request('midas.folder.delete', parameters)
        return response

    def move_folder(self, token, folder_id, dest_folder_id):
        """
        Move a folder to the destination folder.

        :param token: A valid token for the user in question.
        :type token: string
        :param folder_id: The id of the folder to be moved.
        :type folder_id: int | long
        :param dest_folder_id: The id of destination (new parent) folder.
        :type dest_folder_id: int | long
        :returns: Dictionary containing the details of the moved folder.
        :rtype: dict
        """
        parameters = dict()
        parameters['token'] = token
        parameters['id'] = folder_id
        parameters['dstfolderid'] = dest_folder_id
        response = self.request('midas.folder.move', parameters)
        return response

    def create_item(self, token, name, parent_id, **kwargs):
        """
        Create an item to the server.

        :param token: A valid token for the user in question.
        :type token: string
        :param name: The name of the item to be created.
        :type name: string
        :param parent_id: The id of the destination folder.
        :type parent_id: int | long
        :param description: (optional) The description text of the item.
        :type description: string
        :param uuid: (optional) The UUID for the item. It will be generated if
            not given.
        :type uuid: string
        :param privacy: (optional) The privacy state of the item
            ('Public' or 'Private').
        :type privacy: string
        :returns: Dictionary containing the details of the created item.
        :rtype: dict
        """
        parameters = dict()
        parameters['token'] = token
        parameters['name'] = name
        parameters['parentid'] = parent_id
        optional_keys = ['description', 'uuid', 'privacy']
        for key in optional_keys:
            if key in kwargs:
                parameters[key] = kwargs[key]
        response = self.request('midas.item.create', parameters)
        return response

    def item_get(self, token, item_id):
        """
        Get the attributes of the specified item.

        :param token: A valid token for the user in question.
        :type token: string
        :param item_id: The id of the requested item.
        :type item_id: int | string
        :returns: Dictionary of the item attributes.
        :rtype: dict
        """
        parameters = dict()
        parameters['token'] = token
        parameters['id'] = item_id
        response = self.request('midas.item.get', parameters)
        return response

    def download_item(self, item_id, token=None, revision=None):
        """
        Download an item to disk.

        :param item_id: The id of the item to be downloaded.
        :type item_id: int | long
        :param token: (optional) The authentication token of the user
            requesting the download.
        :type token: None | string
        :param revision: (optional) The revision of the item to download, this
            defaults to HEAD.
        :type revision: None | int | long
        :returns: A tuple of the filename and the content iterator.
        :rtype: (string, unknown)
        """
        parameters = dict()
        parameters['id'] = item_id
        if token:
            parameters['token'] = token
        if revision:
            parameters['revision'] = revision
        method_url = self.full_url + 'midas.item.download'
        request = requests.get(method_url,
                               params=parameters,
                               verify=self._verify_ssl_certificate)
        filename = request.headers['content-disposition'][21:].strip('"')
        return filename, request.iter_content(chunk_size=10 * 1024)

    def delete_item(self, token, item_id):
        """
        Delete the item with the passed in item_id.

        :param token: A valid token for the user in question.
        :type token: string
        :param item_id: The id of the item to be deleted.
        :type item_id: int | long
        :returns: None.
        :rtype: None
        """
        parameters = dict()
        parameters['token'] = token
        parameters['id'] = item_id
        response = self.request('midas.item.delete', parameters)
        return response

    def get_item_metadata(self, item_id, token=None, revision=None):
        """
        Get the metadata associated with an item.

        :param item_id: The id of the item for which metadata will be returned
        :type item_id: int | long
        :param token: (optional) A valid token for the user in question.
        :type token: None | string
        :param revision: (optional) Revision of the item. Defaults to latest
            revision.
        :type revision: int | long
        :returns: List of dictionaries containing item metadata.
        :rtype: list[dict]
        """
        parameters = dict()
        parameters['id'] = item_id
        if token:
            parameters['token'] = token
        if revision:
            parameters['revision'] = revision
        response = self.request('midas.item.getmetadata', parameters)
        return response

    def set_item_metadata(self, token, item_id, element, value,
                          qualifier=None):
        """
        Set the metadata associated with an item.

        :param token: A valid token for the user in question.
        :type token: string
        :param item_id: The id of the item for which metadata will be set.
        :type item_id: int | long
        :param element: The metadata element name.
        :type element: string
        :param value: The metadata value for the field.
        :type value: string
        :param qualifier: (optional) The metadata qualifier. Defaults to empty
            string.
        :type qualifier: None | string
        :returns: None.
        :rtype: None
        """
        parameters = dict()
        parameters['token'] = token
        parameters['itemId'] = item_id
        parameters['element'] = element
        parameters['value'] = value
        if qualifier:
            parameters['qualifier'] = qualifier
        response = self.request('midas.item.setmetadata', parameters)
        return response

    def share_item(self, token, item_id, dest_folder_id):
        """
        Share an item to the destination folder.

        :param token: A valid token for the user in question.
        :type token: string
        :param item_id: The id of the item to be shared.
        :type item_id: int | long
        :param dest_folder_id: The id of destination folder where the item is
            shared to.
        :type dest_folder_id: int | long
        :returns: Dictionary containing the details of the shared item.
        :rtype: dict
        """
        parameters = dict()
        parameters['token'] = token
        parameters['id'] = item_id
        parameters['dstfolderid'] = dest_folder_id
        response = self.request('midas.item.share', parameters)
        return response

    def move_item(self, token, item_id, src_folder_id, dest_folder_id):
        """
        Move an item from the source folder to the destination folder.

        :param token: A valid token for the user in question.
        :type token: string
        :param item_id: The id of the item to be moved
        :type item_id: int | long
        :param src_folder_id: The id of source folder where the item is located
        :type src_folder_id: int | long
        :param dest_folder_id: The id of destination folder where the item is
            moved to
        :type dest_folder_id: int | long
        :returns: Dictionary containing the details of the moved item
        :rtype: dict
        """
        parameters = dict()
        parameters['token'] = token
        parameters['id'] = item_id
        parameters['srcfolderid'] = src_folder_id
        parameters['dstfolderid'] = dest_folder_id
        response = self.request('midas.item.move', parameters)
        return response

    def search_item_by_name(self, name, token=None):
        """
        Return all items.

        :param name: The name of the item to search by.
        :type name: string
        :param token: (optional) A valid token for the user in question.
        :type token: None | string
        :returns: A list of all items with the given name.
        :rtype: list[dict]
        """
        parameters = dict()
        parameters['name'] = name
        if token:
            parameters['token'] = token
        response = self.request('midas.item.searchbyname', parameters)
        return response['items']

    def search_item_by_name_and_folder(self, name, folder_id, token=None):
        """
        Return all items with a given name and parent folder id.

        :param name: The name of the item to search by.
        :type name: string
        :param folder_id: The id of the parent folder to search by.
        :type folder_id: int | long
        :param token: (optional) A valid token for the user in question.
        :type token: None | string
        :returns: A list of all items with the given name and parent folder id.
        :rtype: list[dict]
        """
        parameters = dict()
        parameters['name'] = name
        parameters['folderId'] = folder_id
        if token:
            parameters['token'] = token
        response = self.request('midas.item.searchbynameandfolder', parameters)
        return response['items']

    def search_item_by_name_and_folder_name(self, name, folder_name,
                                            token=None):
        """
        Return all items with a given name and parent folder name.

        :param name: The name of the item to search by.
        :type name: string
        :param folder_name: The name of the parent folder to search by.
        :type folder_name: string
        :param token: (optional) A valid token for the user in question.
        :type token: None | string
        :returns: A list of all items with the given name and parent folder
            name.
        :rtype: list[dict]
        """
        parameters = dict()
        parameters['name'] = name
        parameters['folderName'] = folder_name
        if token:
            parameters['token'] = token
        response = self.request('midas.item.searchbynameandfoldername',
                                parameters)
        return response['items']

    def create_link(self, token, folder_id, url, **kwargs):
        """
        Create a link bitstream.

        :param token: A valid token for the user in question.
        :type token: string
        :param folder_id: The id of the folder in which to create a new item
            that will contain the link. The new item will have the same name as
            the URL unless an item name is supplied.
        :type folder_id: int | long
        :param url: The URL of the link you will create, will be used as the
            name of the bitstream and of the item unless an item name is
            supplied.
        :type url: string
        :param item_name: (optional)  The name of the newly created item, if
            not supplied, the item will have the same name as the URL.
        :type item_name: string
        :param length: (optional) The length in bytes of the file to which the
            link points.
        :type length: int | long
        :param checksum: (optional) The MD5 checksum of the file to which the
            link points.
        :type checksum: string
        :returns: The item information of the item created.
        :rtype: dict
        """
        parameters = dict()
        parameters['token'] = token
        parameters['folderid'] = folder_id
        parameters['url'] = url
        optional_keys = ['item_name', 'length', 'checksum']
        for key in optional_keys:
            if key in kwargs:
                if key == 'item_name':
                    parameters['itemname'] = kwargs[key]
                    continue
                parameters[key] = kwargs[key]
        response = self.request('midas.link.create', parameters)
        return response

    def generate_upload_token(self, token, item_id, filename, checksum=None):
        """
        Generate a token to use for upload.

        Midas Server uses a individual token for each upload. The token
        corresponds to the file specified and that file only. Passing the MD5
        checksum allows the server to determine if the file is already in the
        asset store.

        If :param:`checksum` is passed and the token returned is blank, the
        server already has this file and there is no need to follow this
        call with a call to `perform_upload`, as the passed in file will have
        been added as a bitstream to the item's latest revision, creating a
        new revision if one doesn't exist.

        :param token: A valid token for the user in question.
        :type token: string
        :param item_id: The id of the item in which to upload the file as a
            bitstream.
        :type item_id: int | long
        :param filename: The name of the file to generate the upload token for.
        :type filename: string
        :param checksum: (optional) The checksum of the file to upload.
        :type checksum: None | string
        :returns: String of the upload token.
        :rtype: string
        """
        parameters = dict()
        parameters['token'] = token
        parameters['itemid'] = item_id
        parameters['filename'] = filename
        if checksum is not None:
            parameters['checksum'] = checksum
        response = self.request('midas.upload.generatetoken', parameters)
        return response['token']

    def perform_upload(self, upload_token, filename, **kwargs):
        """
        Upload a file into a given item (or just to the public folder if the
        item is not specified.

        :param upload_token: The upload token (returned by
            generate_upload_token)
        :type upload_token: string
        :param filename: The upload filename. Also used as the path to the
            file, if 'filepath' is not set.
        :type filename: string
        :param mode: (optional) Stream or multipart. Default is stream.
        :type mode: string
        :param folder_id: (optional) The id of the folder to upload into.
        :type folder_id: int | long
        :param item_id: (optional) If set, will append item ``bitstreams`` to
            the latest revision (or the one set using :param:`revision` ) of
            the existing item.
        :type item_id: int | long
        :param revision: (optional) If set, will add a new file into an
            existing revision. Set this to 'head' to add to the most recent
            revision.
        :type revision: string | int | long
        :param filepath: (optional) The path to the file.
        :type filepath: string
        :param create_additional_revision: (optional) If set, will create a
            new revision in the existing item.
        :type create_additional_revision: bool
        :returns: Dictionary containing the details of the item created or
            changed.
        :rtype: dict
        """
        parameters = dict()
        parameters['uploadtoken'] = upload_token
        parameters['filename'] = filename

        try:
            create_additional_revision = kwargs['create_additional_revision']
        except KeyError:
            create_additional_revision = False

        if not create_additional_revision:
            parameters['revision'] = 'head'
        optional_keys = ['mode', 'folderid', 'item_id', 'itemid', 'revision']
        for key in optional_keys:
            if key in kwargs:
                if key == 'item_id':
                    parameters['itemid'] = kwargs[key]
                    continue
                if key == 'folder_id':
                    parameters['folderid'] = kwargs[key]
                    continue
                parameters[key] = kwargs[key]

        # We may want a different name than path
        file_payload = open(kwargs.get('filepath', filename), 'rb')
        # Arcane getting of the file size using fstat. More details can be
        # found in the python library docs
        parameters['length'] = os.fstat(file_payload.fileno()).st_size

        response = self.request('midas.upload.perform', parameters,
                                file_payload)
        return response

    def search(self, search, token=None):
        """
        Get the resources corresponding to a given query.

        :param search: The search criterion.
        :type search: string
        :param token: (optional) The credentials to use when searching.
        :type token: None | string
        :returns: Dictionary containing the search result. Notable is the
            dictionary item 'results', which is a list of item details.
        :rtype: dict
        """
        parameters = dict()
        parameters['search'] = search
        if token:
            parameters['token'] = token
        response = self.request('midas.resource.search', parameters)
        return response


class BatchmakeDriver(BaseDriver):
    """Driver for the batchmake module API methods."""

    def add_condor_dag(self, token, batchmaketaskid, dagfilename,
                       dagmanoutfilename):
        """
        Add a Condor DAG to the given Batchmake task.

        :param token: A valid token for the user in question.
        :type token: string
        :param batchmaketaskid: id of the Batchmake task for this DAG
        :type batchmaketaskid: int | long
        :param dagfilename: Filename of the DAG file
        :type dagfilename: string
        :param dagmanoutfilename: Filename of the DAG processing output
        :type dagmanoutfilename: string
        :returns: The created Condor DAG DAO
        :rtype: dict
        """
        parameters = dict()
        parameters['token'] = token
        parameters['batchmaketaskid'] = batchmaketaskid
        parameters['dagfilename'] = dagfilename
        parameters['outfilename'] = dagmanoutfilename
        response = self.request('midas.batchmake.add.condor.dag', parameters)
        return response

    def add_condor_job(self, token, batchmaketaskid, jobdefinitionfilename,
                       outputfilename, errorfilename, logfilename,
                       postfilename):
        """
        Add a Condor DAG job to the Condor DAG associated with this
        Batchmake task

        :param token: A valid token for the user in question.
        :type token: string
        :param batchmaketaskid: id of the Batchmake task for this DAG
        :type batchmaketaskid: int | long
        :param jobdefinitionfilename: Filename of the definition file for the
            job
        :type jobdefinitionfilename: string
        :param outputfilename: Filename of the output file for the job
        :type outputfilename: string
        :param errorfilename: Filename of the error file for the job
        :type errorfilename: string
        :param logfilename: Filename of the log file for the job
        :type logfilename: string
        :param postfilename: Filename of the post script log file for the job
        :type postfilename: string
        :return: The created Condor job DAO.
        :rtype: dict
        """
        parameters = dict()
        parameters['token'] = token
        parameters['batchmaketaskid'] = batchmaketaskid
        parameters['jobdefinitionfilename'] = jobdefinitionfilename
        parameters['outputfilename'] = outputfilename
        parameters['errorfilename'] = errorfilename
        parameters['logfilename'] = logfilename
        parameters['postfilename'] = postfilename
        response = self.request('midas.batchmake.add.condor.job', parameters)
        return response


class DicomextractorDriver(BaseDriver):
    """Driver for the dicomextractor module API methods."""

    def extract_dicommetadata(self, token, item_id):
        """
        Extract DICOM metadata from the given item

        :param token: A valid token for the user in question.
        :type token: string
        :param item_id: id of the item to be extracted
        :type item_id: int | long
        :return: the item revision DAO
        :rtype: dict
        """
        parameters = dict()
        parameters['token'] = token
        parameters['item'] = item_id
        response = self.request('midas.dicomextractor.extract', parameters)
        return response


class MultiFactorAuthenticationDriver(BaseDriver):
    """Driver for the multi-factor authentication module API methods."""

    def mfa_otp_login(self, temp_token, one_time_pass):
        """
        Log in to get the real token using the temporary token and otp.

        :param temp_token: The temporary token or id returned from normal login
        :type temp_token: string
        :param one_time_pass: The one-time pass to be sent to the underlying
            multi-factor engine.
        :type one_time_pass: string
        :returns: A standard token for interacting with the web api.
        :rtype: string
        """
        parameters = dict()
        parameters['mfaTokenId'] = temp_token
        parameters['otp'] = one_time_pass
        response = self.request('midas.mfa.otp.login', parameters)
        return response['token']


class ThumbnailCreatorDriver(BaseDriver):
    """Driver for the thumbnailcreator module API methods."""

    def create_big_thumbnail(self, token, bitstream_id, item_id, width=575):
        """
        Create a big thumbnail for the given bitstream with the given width.
        It is used as the main image of the given item and shown in the item
        view page.

        :param token: A valid token for the user in question.
        :type token: string
        :param bitstream_id: The bitstream from which to create the thumbnail.
        :type bitstream_id: int | long
        :param item_id: The item on which to set the thumbnail.
        :type item_id: int | long
        :param width: (optional) The width in pixels to which to resize (aspect
            ratio will be preserved). Defaults to 575.
        :type width: int | long
        :returns: The ItemthumbnailDao object that was created.
        :rtype: dict
        """
        parameters = dict()
        parameters['token'] = token
        parameters['bitstreamId'] = bitstream_id
        parameters['itemId'] = item_id
        parameters['width'] = width
        response = self.request('midas.thumbnailcreator.create.big.thumbnail',
                                parameters)
        return response

    def create_small_thumbnail(self, token, item_id):
        """
        Create a 100x100 small thumbnail for the given item. It is used for
        preview purpose and displayed in the 'preview' and 'thumbnails'
        sidebar sections.

        :param token: A valid token for the user in question.
        :type token: string
        :param item_id: The item on which to set the thumbnail.
        :type item_id: int | long
        :returns: The item object (with the new thumbnail id) and the path
            where the newly created thumbnail is stored.
        :rtype: dict
        """
        parameters = dict()
        parameters['token'] = token
        parameters['itemId'] = item_id
        response = self.request(
            'midas.thumbnailcreator.create.small.thumbnail', parameters)
        return response


class SolrDriver(BaseDriver):
    """Driver for the solr module API methods."""

    def solr_advanced_search(self, query, token=None, limit=20):
        """
        Search item metadata using Apache Solr.

        :param query: The Apache Lucene search query.
        :type query: string
        :param token: (optional) A valid token for the user in question.
        :type token: None | string
        :param limit: (optional) The limit of the search.
        :type limit: int | long
        :returns: The list of items that match the search query.
        :rtype: list[dict]
        """
        parameters = dict()
        parameters['query'] = query
        parameters['limit'] = limit
        if token:
            parameters['token'] = token
        response = self.request('midas.solr.search.advanced', parameters)
        return response


class TrackerDriver(BaseDriver):
    """Driver for the tracker module API methods."""

    def associate_item_with_scalar_data(self, token, item_id, scalar_id,
                                        label):
        """
        Associate a result item with a particular scalar value.

        :param token: A valid token for the user in question.
        :type token: string
        :param item_id: The id of the item to associate with the scalar.
        :type item_id: int | long
        :param scalar_id: Scalar id with which to associate the item.
        :type scalar_id: int | long
        :param label: The label describing the nature of the association.
        :type label: string
        """
        parameters = dict()
        parameters['token'] = token
        parameters['scalarIds'] = scalar_id
        parameters['itemId'] = item_id
        parameters['label'] = label
        self.request('midas.tracker.item.associate', parameters)

    def create_submission(self, token, **kwargs):
        """
        Associate a result item with a particular scalar value.

        :param token: A valid token for the user in question.
        :type token: string
        :param uuid (optional) The uuid of the submission (must be unique)
        :type uuid: string
        :param name (optional) The name of the submission
        :type name: string
        :returns: The submission object that was created.
        :rtype: dict
        """
        parameters = {}
        parameters['token'] = token
        optional_keys = ['uuid', 'name']
        for key in optional_keys:
            if key in kwargs:
                parameters[key] = kwargs[key]
        return self.request('midas.tracker.submission.create', parameters)

    def add_scalar_data(self, token, community_id, producer_display_name,
                        metric_name, producer_revision, submit_time, value,
                        **kwargs):
        """
        Create a new scalar data point.

        :param token: A valid token for the user in question.
        :type token: string
        :param community_id: The id of the community that owns the producer.
        :type community_id: int | long
        :param producer_display_name: The display name of the producer.
        :type producer_display_name: string
        :param metric_name: The metric name that identifies which trend this
            point belongs to.
        :type metric_name: string
        :param producer_revision: The repository revision of the producer that
            produced this value.
        :type producer_revision: int | long | string
        :param submit_time: The submit timestamp. Must be parsable with PHP
            strtotime().
        :type submit_time: string
        :param value: The value of the scalar.
        :type value: float
        :param config_item_id: (optional) If this value pertains to a specific
            configuration item, pass its id here.
        :type config_item_id: int | long
        :param test_dataset_id: (optional) If this value pertains to a
            specific test dataset, pass its id here.
        :type test_dataset_id: int | long
        :param truth_dataset_id: (optional) If this value pertains to a
            specific ground truth dataset, pass its id here.
        :type truth_dataset_id: int | long
        :param silent: (optional) If true, do not perform threshold-based email
            notifications for this scalar.
        :type silent: bool
        :param unofficial: (optional) If true, creates an unofficial scalar
            visible only to the user performing the submission.
        :type unofficial: bool
        :param build_results_url: (optional) A URL for linking to build results
            for this submission.
        :type build_results_url: string
        :param branch: (optional) The branch name in the source repository for
            this submission.
        :type branch: string
        :param submission_id: (optional) The id of the submission.
        :type submission_id: int | long
        :param submission_uuid: (optional) The uuid of the submission. If one
            does not exist, it will be created.
        :type submission_uuid: string
        :type branch: string
        :param params: (optional) Any key/value pairs that should be displayed
            with this scalar result.
        :type params: dict
        :param extra_urls: (optional) Other URL's that should be displayed with
            with this scalar result. Each element of the list should be a dict
            with the following keys: label, text, href
        :type extra_urls: list[dict]
        :param unit: (optional) The unit of the scalar value.
        :type unit: string
        :returns: The scalar object that was created.
        :rtype: dict
        """
        parameters = dict()
        parameters['token'] = token
        parameters['communityId'] = community_id
        parameters['producerDisplayName'] = producer_display_name
        parameters['metricName'] = metric_name
        parameters['producerRevision'] = producer_revision
        parameters['submitTime'] = submit_time
        parameters['value'] = value
        optional_keys = [
            'config_item_id', 'test_dataset_id', 'truth_dataset_id', 'silent',
            'unofficial', 'build_results_url', 'branch', 'extra_urls',
            'params', 'submission_id', 'submission_uuid', 'unit']
        for key in optional_keys:
            if key in kwargs:
                if key == 'config_item_id':
                    parameters['configItemId'] = kwargs[key]
                elif key == 'test_dataset_id':
                    parameters['testDatasetId'] = kwargs[key]
                elif key == 'truth_dataset_id':
                    parameters['truthDatasetId'] = kwargs[key]
                elif key == 'build_results_url':
                    parameters['buildResultsUrl'] = kwargs[key]
                elif key == 'extra_urls':
                    parameters['extraUrls'] = json.dumps(kwargs[key])
                elif key == 'params':
                    parameters[key] = json.dumps(kwargs[key])
                elif key == 'silent':
                    if kwargs[key]:
                        parameters[key] = kwargs[key]
                elif key == 'unofficial':
                    if kwargs[key]:
                        parameters[key] = kwargs[key]
                elif key == 'submission_id':
                    parameters['submissionId'] = kwargs[key]
                elif key == 'submission_uuid':
                    parameters['submissionUuid'] = kwargs[key]
                elif key == 'unit':
                    parameters['unit'] = kwargs[key]
                else:
                    parameters[key] = kwargs[key]
        response = self.request('midas.tracker.scalar.add', parameters)
        return response

    def upload_json_results(self, token, filepath, community_id,
                            producer_display_name, metric_name,
                            producer_revision, submit_time, **kwargs):
        """
        Upload a JSON file containing numeric scoring results to be added as
        scalars. File is parsed and then deleted from the server.

        :param token: A valid token for the user in question.
        :param filepath: The path to the JSON file.
        :param community_id: The id of the community that owns the producer.
        :param producer_display_name: The display name of the producer.
        :param producer_revision: The repository revision of the producer
            that produced this value.
        :param submit_time: The submit timestamp. Must be parsable with PHP
            strtotime().
        :param config_item_id: (optional) If this value pertains to a specific
            configuration item, pass its id here.
        :param test_dataset_id: (optional) If this value pertains to a
            specific test dataset, pass its id here.
        :param truth_dataset_id: (optional) If this value pertains to a
            specific ground truth dataset, pass its id here.
        :param parent_keys: (optional) Semicolon-separated list of parent keys
            to look for numeric results under. Use '.' to denote nesting, like
            in normal javascript syntax.
        :param silent: (optional) If true, do not perform threshold-based email
            notifications for this scalar.
        :param unofficial: (optional) If true, creates an unofficial scalar
            visible only to the user performing the submission.
        :param build_results_url: (optional) A URL for linking to build results
            for this submission.
        :param branch: (optional) The branch name in the source repository for
            this submission.
        :param params: (optional) Any key/value pairs that should be displayed
            with this scalar result.
        :type params: dict
        :param extra_urls: (optional) Other URL's that should be displayed with
            with this scalar result. Each element of the list should be a dict
            with the following keys: label, text, href
        :type extra_urls: list of dicts
        :returns: The list of scalars that were created.
        """
        parameters = dict()
        parameters['token'] = token
        parameters['communityId'] = community_id
        parameters['producerDisplayName'] = producer_display_name
        parameters['metricName'] = metric_name
        parameters['producerRevision'] = producer_revision
        parameters['submitTime'] = submit_time
        optional_keys = [
            'config_item_id', 'test_dataset_id', 'truth_dataset_id', 'silent',
            'unofficial', 'build_results_url', 'branch', 'extra_urls',
            'params']
        for key in optional_keys:
            if key in kwargs:
                if key == 'config_item_id':
                    parameters['configItemId'] = kwargs[key]
                elif key == 'test_dataset_id':
                    parameters['testDatasetId'] = kwargs[key]
                elif key == 'truth_dataset_id':
                    parameters['truthDatasetId'] = kwargs[key]
                elif key == 'parent_keys':
                    parameters['parentKeys'] = kwargs[key]
                elif key == 'build_results_url':
                    parameters['buildResultsUrl'] = kwargs[key]
                elif key == 'extra_urls':
                    parameters['extraUrls'] = json.dumps(kwargs[key])
                elif key == 'params':
                    parameters[key] = json.dumps(kwargs[key])
                elif key == 'silent':
                    if kwargs[key]:
                        parameters[key] = kwargs[key]
                elif key == 'unofficial':
                    if kwargs[key]:
                        parameters[key] = kwargs[key]
                else:
                    parameters[key] = kwargs[key]
        file_payload = open(filepath, 'rb')
        response = self.request('midas.tracker.results.upload.json',
                                parameters, file_payload)
        return response
